
//
// StarChart.java: A quick star-charting Java class.
//	Copyright 1996 by Akkana Peck.
//

package Akk.Astro;

import java.awt.*;
import java.io.*;
import java.net.*;
import java.applet.*;

class DrawnStar
{
    final static int Unknown = 0;
    final static int Star = 1;
    final static int Nebula = 2;
    final static int Planetoid = 3;
    final static int Line = 4;		// draw a line from the last coord
    final static int Label = 5;		// e.g. name for a constellation

    static int lastX, lastY;

    int type;			// Star, Nebula or Planetoid
    int x, y;
    float mag;
    String name;
    AstroCoords coords;		// RA and Dec

    DrawnStar(int t, int xx, int yy, float mmag, String n) {
	type=t; x=xx; y=yy; mag=mmag; name=n;
    }

    DrawnStar(int t, int xx, int yy, float mmag, String n, AstroCoords ac) {
	type=t; x=xx; y=yy; mag=mmag; name=n; coords = ac;
    }

    void draw(Graphics g, StarChart chart)
    {
	// Assume we're set to chart.starcolor already
	if (type == Star) {
	    // Make sure this star is bright enough to display:
	    if (mag > chart.getStarMagnitude())
		return;

	    // only name the brightest
	    if (name != null && mag < 1.75 && y > 20)
		g.drawString(name, x+4, y+6);

	    if (mag <= 0) {
		g.fillRect(x-1, y-1, 4, 4);
	    } else if (mag <= 3.5) {
		g.fillRect(x-1, y-1, 3, 3);
	    } else if (mag < 4.8) {
		g.drawLine(x, y, x+1, y);
		g.drawLine(x, y+1, x+1, y+1);
	    } else if (mag < 6) {
		g.drawLine(x, y, x+1, y);
	    } else {
		g.drawLine(x, y, x, y);
	    }

	} else if (type == Nebula) {
	    g.setColor(chart.nebcolor);
	    g.drawLine(x-1, y, x-1, y);
	    g.drawLine(x, y+1, x, y+1);
	    g.drawLine(x+1, y, x+1, y);
	    g.drawLine(x, y-1, x, y-1);
	    g.drawString(name, x+4, y+6);
	    g.setColor(chart.starcolor);

	} else if (type == Planetoid) {
	    g.setColor(chart.highlightColor);
	    g.drawString("* "+name, x+4, y+6);
	    g.setColor(chart.starcolor);
	} else if (type == Line) {
	    g.drawLine(lastX, lastY, x, y);
	} else if (type == Label) {
	    g.setColor(chart.starcolor);
	    g.drawString(name, x, y);
	}
	lastX = x;	lastY = y;
    }
};

public class StarChart extends Canvas implements Runnable
{
    //
    // Limits of the chart:
    //
    double RA, Dec;
    double RAspan;
    double decspan;
    AstroCoords limit1, limit2;
    StarDate now;
    int numDays;
    double starMagnitude;	// min. magnitude to display
    double loadMagnitude;	// min. magnitude of stars currently loaded

    // colors
    boolean inverted;
    Color background;
    Color starcolor;
    Color nebcolor;
    Color linecolor;
    Color highlightColor;

    // the thread in which the star catalog will be loaded:
    Thread loadThread;

    // a place to print our messages.  If null,
    // messages will be printed on the graphics panel itself.
    Label mesgLabl;

    // The URL from which to read the star catalog
    String StarURLname;

    // place to save all the info once it's calculated, for fast redraws:
    // maxNumPlanets is actually the max number of planet points shown,
    // i.e. # planets * # times calculated:
    final int maxNumPlanets = 100;
    int numPlanets;
    DrawnStar drawnPlanets[];

    final int maxNumStars = 5000;
    volatile int numStars;	// updated by star-loading thread
    DrawnStar drawnStars[];

    // flags indicating whether we've done the calculations yet
    boolean loadedStars;
    boolean calculatedPlanets;

    // Size as set from the parent calling resize():
    int preferredWidth;
    int preferredHeight;

    // Used by showMessage:
    String lastMessage = null;

    static Font starNameFont;
    static Font messageFont;

    //
    // Constructor
    //
    public StarChart(double _RA, double _Dec, double _decspan,
		     String _starURL, int _numDays, double smag)
    {
	// Copy the args which don't need additional calculation:
	RA = _RA;
	Dec = _Dec;
	decspan = _decspan;
	numDays = _numDays;
	StarURLname = _starURL;
	starMagnitude = smag;
	loadMagnitude = smag;

	now = new StarDate();		// use the current time

	// Initialize some variables that we'll need later:
	loadedStars = false;
	numStars = 0;
	calculatedPlanets = false;
	numPlanets = 0;

	// Would be nice to be able to grow this dynamically
	// (we can do that if we use the Vector class ...)
	drawnPlanets = new DrawnStar[maxNumPlanets];
	drawnStars = new DrawnStar[maxNumStars];
    }

    //
    // Public methods
    //
    int getWidth() { return size().width; }
    int getHeight() { return size().height; }

    // Have to recalculate limits each time we're resized:
    void calcLimits() { calcLimits(getWidth(), getHeight()); }
    void calcLimits(int w, int h)
    {
	// Calculate limits:
	RAspan = decspan * w / h * 12 / 180;
	double RA1 = RA - RAspan / 2;
	double RA2 = RA + RAspan / 2;
	// See if RA2 wraps; if so, mod it for better performance in onChart:
	if(RA2 > 24)
	    RA2 -= 24;
	double dec1 = Dec - decspan / 2;
	double dec2 = Dec + decspan / 2;

	// And turn them into a pair of AstroCoords for better
	// performance in onChart():
	limit1 = new AstroCoords(RA1 * Math.PI / 12, dec1 * Math.PI / 180);
	limit2 = new AstroCoords(RA2 * Math.PI / 12, dec2 * Math.PI / 180);
    }
    public void resize(int w, int h)
    {
	preferredWidth = w;
	preferredHeight = h;
	super.resize(w, h);
	calcLimits(w, h);
	loadedStars = false;		// re-evaluate: not always necessary
	calculatedPlanets = false;
    }
    public void resize(Dimension d)
    {
	super.resize(d);
	calcLimits(d.width, d.height);
    }

    // swap foreground and background colors
    public void invert() {
	invert(inverted = !inverted);
    }

    public void invert(boolean b)
    {
	inverted = b;
	if (inverted) {
	    background = Color.white;
	    starcolor = Color.black;
	    linecolor = Color.gray;
	    nebcolor = Color.red;
	    highlightColor = Color.blue;
	} else {
	    background = Color.black;
	    starcolor = Color.white;
	    linecolor = Color.gray;
	    nebcolor = Color.green;
	    highlightColor = Color.yellow;
	}
	setBackground(background);
	setForeground(starcolor);
    }

    public void setMessages(Label l)	{ mesgLabl = l; }

    public StarDate getDate()	{ return now; }
    public void setDate(StarDate newdate) {
	calculatedPlanets = false;
	now = newdate;
    }
    public int getNumDays()	{ return numDays; }
    public void setNumDays(int nd) {
	calculatedPlanets = false;
	numDays = nd;
    }

    public double getRA()	{ return RA; }
    public double getDec()	{ return Dec; }
    public double getDecSpan()	{ return decspan; }
    public void setCoordinates(double _RA, double _Dec, double _decspan)
    {
	// If nothing has changed, then don't do anything:
	if (_RA == RA && _Dec == Dec && _decspan == decspan)
	    return;

	// save old limits, to see later whether we need to reload stars:
	AstroCoords oldLimit1 = new AstroCoords(limit1.get_RA_radians(),
						limit1.get_dec_radians());
	AstroCoords oldLimit2 = new AstroCoords(limit2.get_RA_radians(),
						limit2.get_dec_radians());
	double oldDecspan = decspan;
	double oldRAspan = RAspan;

	// calc new limits:
	RA = _RA;
	Dec = _Dec;
	decspan = _decspan;
	calcLimits();

	// See whether we really need to reload stars:
	if (limit1.get_RA_radians() < oldLimit1.get_RA_radians() ||
	    limit2.get_RA_radians() > oldLimit2.get_RA_radians() ||
	    limit1.get_dec_radians() < oldLimit1.get_dec_radians() ||
	    limit2.get_dec_radians() > oldLimit2.get_dec_radians())
	{
	    loadedStars = false;
	    calculatedPlanets = false;

	    // If we're going to reload stars, then we'd better make
	    // sure there's no current load already going on:
	    if (loadThread != null && loadThread.isAlive()) {
		loadThread.stop();
	    }
	} else {
	    boolean suspended = false;
	    String saveMesg = lastMessage;
	    if (loadThread != null && loadThread.isAlive()) {
		suspended = true;

		// suspending the load thread isn't enough -- we also
		// have to yield() in order to get any time ourselves
		Thread.yield();
		loadThread.suspend();
	    }

	    rescaleStars(oldLimit1, oldDecspan, oldRAspan);

	    if (suspended) {
		Thread.yield();
		loadThread.resume();
		showMessage(saveMesg);
	    }
	}
    }

    public double getStarMagnitude()	{ return starMagnitude; }
    public double getLoadMagnitude()	{ return loadMagnitude; }
    public void setStarMagnitude(double m) {
	if (m > loadMagnitude) {
	    loadMagnitude = m;
	    loadedStars = false;
	}
	starMagnitude = m;
    }

    // Tell layout managers that we want to be as big as possible:
    public Dimension preferredSize() {
	return new Dimension(getParent().size().width, preferredHeight);
    }
    public Dimension minimumSize() {
	return new Dimension(getParent().size().width, 10);
    }

    // Laura Lemay says that the defaul update() clears the screen by
    // default, but maybe she only means for Applets.  Empirically, it
    // does not for this class, and so needs to be redefined to ensure
    // that old stars get erased when reloading or rescaling:
    public void update(Graphics g)
    {
	g.setColor(background);
	g.fillRect(0, 0, getWidth(), getHeight());
	paint(g);
    }

    public void paint(Graphics g)
    {
	// Initialize the colors the first time around.
	// If we do this in the ctor, it causes a thread conflict
	// which sometimes makes appletviewer dump core.
	if (background == null)
	    invert(false);

	if (starNameFont == null)
	    starNameFont = new Font("TimesRoman", Font.ITALIC, 10);
	g.setFont(starNameFont);
	g.setColor(starcolor);
	drawGrid(g);

	// Should spawn a new thread to calculate planet and star positions!

	if (!calculatedPlanets)
	    calcPlanets();
	for (int j=0; j < numPlanets; ++j) {
	    drawnPlanets[j].draw(g, this);
	}

	// read in and draw the stars
	if (!loadedStars)
	    loadStars();

	for (int j=0; j < numStars; ++j)
	    drawnStars[j].draw(g, this);
    }

    //
    // End of public interface
    //

    //
    // Useful conversion routines
    //
    int astroCoords2X(AstroCoords ac) {
	if (limit1.get_RA_hours() > limit2.get_RA_hours())  // we're wrapping
	{
	    if (ac.get_RA_hours() < limit2.get_RA_hours())  // in "normal" half
		return (int)(getWidth() * (1 - ((ac.get_RA_hours()
						 - limit1.get_RA_hours() + 24)
						/ RAspan)));
	}

        return (int)(getWidth() * (1 - ((ac.get_RA_hours()
					 - limit1.get_RA_hours())
					/ RAspan)));
    }

    int astroCoords2Y(AstroCoords ac) {
        return (int)(getHeight() *
		     (1 - ((ac.get_dec_degrees() - limit1.get_dec_degrees())
			   / decspan)));
    }

    boolean onChart(AstroCoords ac) {
	return ac.inside(limit1, limit2);
    }

    // The only record we have of the stars we loaded is in
    // the drawnStars[numStars] array.  In order to avoid
    // reloading, we have to rescale these.
    // While we're at it, we do the planets as well.
    void rescaleStars(AstroCoords oldLimit1,
		      double oldDecspan, double oldRAspan)
    {
	showMessage("rescaling stars ...");

	int j;
	for (j=0; j < numPlanets; ++j) {
	    double new_ra = oldLimit1.get_RA_hours() + oldRAspan *
		(1-(double)drawnPlanets[j].x/getWidth());
	    double new_dec = oldLimit1.get_dec_degrees() + oldDecspan *
		(1-(double)drawnPlanets[j].y/getHeight());
	    AstroCoords newcoords = new AstroCoords(new_ra * Math.PI / 12,
						    new_dec * Math.PI / 180);

	    drawnPlanets[j].x = astroCoords2X(newcoords);
	    drawnPlanets[j].y = astroCoords2Y(newcoords);
	}
	for (j=0; j < numStars; ++j) {
	    double new_ra = oldLimit1.get_RA_hours() + oldRAspan *
		(1. - (double)drawnStars[j].x/getWidth());
	    double new_dec = oldLimit1.get_dec_degrees() + oldDecspan *
		(1 - (double)drawnStars[j].y/getHeight());
	    AstroCoords newcoords = new AstroCoords(new_ra * Math.PI / 12,
						    new_dec * Math.PI / 180);
	    drawnStars[j].x = astroCoords2X(newcoords);
	    drawnStars[j].y = astroCoords2Y(newcoords);
	}

	showMessage(null);
    }

    //
    // Find positions for moving objects (planets)
    //
    void calcPlanets()
    {
	calculatedPlanets = true;

	numPlanets = 0;

	StarDate thistime = new StarDate(now);
	int increment = 1;
	// If we're showing planets over a larger number of days,
	// then use a bigger increment, and make sure that we have
	// enough space in the drawnPlanets[] array to hold everything:
	if (numDays > 4) {
	    increment = numDays/3;
	    while ((numDays / increment * Planet.knownObjects.length
		    > drawnPlanets.length)
		   && increment < numDays)
		increment += 1;
	}

	try {
	    for (int day=0; day < numDays; day += increment)
	    {
		// Calculate the positions of the known objects:
		// Start by calculating the earth's heliocentric coords:
		HelioCoords earthc = Planet.earth.calcHeliocentric(thistime);

		// then use that to calculate positions for each object:
		for (int i = 0; i < Planet.knownObjects.length; ++i)
		{
		    Planet obj = Planet.knownObjects[i];

		    AstroCoords coords = obj.locate(thistime, earthc);

		    // See if it's on our chart:
		    if (!onChart(coords))
			continue;

		    // save it in the array:
		    int x = astroCoords2X(coords);
		    int y = astroCoords2Y(coords);
		    if (day == 0)
			drawnPlanets[numPlanets++] =
			new DrawnStar(DrawnStar.Planetoid, x, y,
				      (float)coords.dist,
				      obj.name, coords);
		    else
			drawnPlanets[numPlanets++] =
	                new DrawnStar(DrawnStar.Planetoid, x, y,
				      (float)coords.dist,
				      "", coords);
		    // would prefer to draw lines, but I'll add that later:
		    // it's harder with the date loop outside the planet loop
		}

		thistime.addDays(increment);
	    }
	} // end "try"
	catch (ArrayIndexOutOfBoundsException e) {
	    showMessage("Too many planets!");
	}
    }

    //
    // Load the star and nebula databases
    // (slow, but only needed when coords change)
    //
    void loadStars()
    {
	if (loadThread != null && loadThread.isAlive()) {
	    return;
	}

	if (loadedStars)
	    return;
	loadedStars = true;

	// Otherwise, we have to destroy the old load thread
	// and create a new one -- calling start() a second time
	// on an old thread doesn't work (the thread never runs).
	loadThread = new Thread(this, "star chart loader");

	// Don't do this: it makes the priority too low to run.
	// What is the main thread doing during this time?
	//loadThread.setPriority(Thread.MIN_PRIORITY);

	loadThread.start();
    }

    //
    // run() is only run from the loadThread,
    // to load the star atlas in a separate thread.
    //
    public void run()
    {
	numStars = 0;

	// Load the constellation names:
        for (int i = 0 ; i < constellations.length; ++i)
	{
	    AstroCoords ac = new AstroCoords(constellations[i][1],
					     constellations[i][2]);

	    // See if it's on our chart:
	    if (!onChart(ac))
		continue;

	    // calculate the position on our chart:
	    int x = astroCoords2X(ac);
	    int y = astroCoords2Y(ac);

	    // and save it in the array:
	    drawnStars[numStars++] = new DrawnStar(DrawnStar.Label,
						   x, y, 0,
						   constellations[i][0],
						   ac);
	}

	showMessage("Reading nebulae ...");

    	// Loop over the nebula catalog:
        for (int i = 0 ; i < nebulae.length; ++i)
	{
	    AstroCoords ac = new AstroCoords(nebulae[i][1],
					     nebulae[i][2]);

	    // See if it's on our chart:
	    if (!onChart(ac))
		continue;

	    // Get the magnitude:
	    float mag = Float.valueOf(nebulae[i][3]).floatValue();
	    // and see if we're supposed to display it -- show nebulae
	    // somewhat fainter than the stars we're showing:
	    if (mag > loadMagnitude + 1.5)
	    	continue;

	    // calculate the position on our chart:
	    int x = astroCoords2X(ac);
	    int y = astroCoords2Y(ac);

	    // and save it in the array:
	    drawnStars[numStars++] = new DrawnStar(DrawnStar.Nebula,
						   x, y, mag,
						   nebulae[i][0],
						   ac);
	}

	repaint();
	showMessage("Connecting: " + StarURLname);

	try {
	    // Append the current coordinates in CGI format, e.g.
	    // ${StarURLname}?RAmin=17&RAmax=21&Decmin=-35&Decmax=10
	    String urlname = StarURLname;
	    if (urlname.indexOf(".cgi") >= 0) {
		urlname += "?RAmin=";
		urlname += new Integer((int)limit1.get_RA_hours()).toString();
		urlname += "&RAmax=";
		urlname += new Integer((int)limit2.get_RA_hours()+1).toString();
		urlname += "&Decmin=";
		urlname += new Integer((int)limit1.get_dec_degrees()).toString();
		urlname += "&Decmax=";
		urlname += new Integer((int)limit2.get_dec_degrees()+1).toString();
		urlname += "&mag=";
		urlname += new Double(loadMagnitude).toString();
	    }

	    // Get star list from a URL:
	    URL StarURL = new URL(urlname);

	    URLConnection conn = StarURL.openConnection();
	    conn.connect();
	    DataInputStream starin
		= new DataInputStream(conn.getInputStream());

	    showMessage("Reading stars: " + StarURLname);
	    while (true)
	    {
		String line = starin.readLine();
		if (line == null) break;

		// Parse the three important parameters from the line.
		// A typical line:
		// <tr><td>06h22m57.6s</td><td>+22^30'49"</td><td> 2.88</td></tr>
		AstroCoords ac = new AstroCoords(0, 0);
		float mag;
		int index;
		index = line.indexOf("<td>");
		if (index < 0)
		    continue;

		index += 4;
		ac.setRAFromString(line.substring(index));
		index = line.indexOf("<td>", index);
		if (index < 0)
		    continue;
		index += 4;
		ac.setDecFromString(line.substring(index));

		if (!onChart(ac))
		    continue;

		index = line.indexOf("<td>", index);
		index += 4;
		if (index < 0)
		    continue;

		mag = Float.valueOf(line.substring(index)).floatValue();
		if (mag > loadMagnitude)
		    break;

		// calculate the position on our chart:
		int x = astroCoords2X(ac);
		int y = astroCoords2Y(ac);

		// and draw it then save it in the array:
		DrawnStar ds = new DrawnStar(DrawnStar.Star,
					     x, y, mag,
					     (String)null);
		// Note that we don't actually save the AstroCoords ac,
		// since for stars we'll never use it, so why allocate
		// the extra object?
		drawnStars[numStars++] = ds;

		// I can't seem to find any way to get this star to become
		// visible here!  Even yielding the thread doesn't help,
		// yet somehow, if we call paint(), the main thread then
		// gets enough cycles to redraw everything.
		//ds.draw(getGraphics(), this);
		//Thread.yield();

		// So as a temporary solution, we periodically call paint().
		if ((numStars % 100) == 0) {
		    showMessage(new Integer(numStars).toString()
				+ " stars loaded");
		    paint(getGraphics());

		    // However, we still have to yield occasionally,
		    // otherwise we can't act on anything the user does
		    // via the UI while we're loading.
		    Thread.yield();
		}
	    }
	    showMessage(null);
	}

	catch (ArrayIndexOutOfBoundsException e) {
	    numStars = drawnStars.length;
	    showMessage("Too many stars!");
	}
	catch (MalformedURLException e)	{
	    showMessage(lastMessage + "\n -- Malformed URL");
	}
	catch (IOException e) {
	    showMessage(lastMessage + "\n -- IO Exception");
	}
	finally {
	    repaint();
	}
    }

    void drawGrid(Graphics g)
    {
	// Draw a line at every 10 degrees of declination:
	for (int thisDec = (int)(limit1.get_dec_degrees() / 10) * 10 - 10;
	     thisDec < limit2.get_dec_degrees() + 10;
	     thisDec += 10)
	{
	    AstroCoords ac = new AstroCoords(0, (double)thisDec * Math.PI/180);
	    int thisY = astroCoords2Y(ac);
	    g.setColor(linecolor);
	    g.drawLine(0, thisY, getWidth(), thisY);
	    g.setColor(starcolor);
	    g.drawString(new Integer(thisDec).toString(), 10, thisY);
	}

	int thisRA = (int)limit1.get_RA_hours();
	int finalRA = (int)limit2.get_RA_hours() + 1;
	if (thisRA > finalRA)
	    thisRA -= 24;
	thisRA -= 2;
	// and at every one-hour interval:
	while (thisRA++ < finalRA)
	{
	    AstroCoords ac = new AstroCoords((double)thisRA * Math.PI/12, 0);
	    int thisX = astroCoords2X(ac);
	    g.setColor(linecolor);
	    g.drawLine(thisX, 0, thisX, getHeight());
	    g.setColor(starcolor);
	    g.drawString(new Integer((thisRA+24) % 24).toString(),
			 thisX+5, getHeight()-5);
	}
    }

    void showMessage(String message)  // Show the message
    {
	if (mesgLabl != null) {
	    mesgLabl.setText(message);

	} else {
	    if (messageFont == null)
		messageFont = new Font("TimesRoman", Font.PLAIN, 12);

	    Graphics g = getGraphics();
	    //g.setXORMode(background);

	    // erase the last message:
	    if (lastMessage != null) {
		g.setFont(messageFont);
		g.drawString(lastMessage, 5, 10);
	    }

	    if (message != null) {
		g.setFont(messageFont);
		g.drawString(message, 5, 10);
	    }

	    g.setFont(starNameFont);
	}

	//System.out.println("[ " + message + " ]");
	lastMessage = message;
    }

    // When the mouse button is clicked, display the name and coords
    // of the most interesting object closest to the mouse.
    public boolean mouseDown(Event event, int x, int y)
    {
	double dist = Math.sqrt(getHeight()*getHeight()
				+ getWidth()*getWidth());
	DrawnStar obj = null;
	boolean wasPlanet = true;
	int threshold = 70;

	// First, look for solar system objects:
	for (int i=0; i < numPlanets; ++i) {
	    double newdist = Math.sqrt((drawnPlanets[i].x - x)
				       * (drawnPlanets[i].x - x)
				       + (drawnPlanets[i].y - y)
				        * (drawnPlanets[i].y - y));
	    if (newdist < dist)
	    {
		dist = newdist;
		obj = drawnPlanets[i];
		wasPlanet = true;
	    }
	}

	// Now, look for nebulae:  they're all clustered at the beginning
	// of drawnStars[numStars].
	for (int i=0; i < numStars; ++i) {
	    if (drawnStars[i].type == DrawnStar.Star)
		break;

	    int newdist = Math.abs(drawnStars[i].x - x)
		+ Math.abs(drawnStars[i].y - y);
	    if (newdist < dist) {
		dist = newdist;
		obj = drawnStars[i];
		wasPlanet = false;
	    }
	}

	// See if it's "too far away" -- if it is, don't show anything
	if (dist > 10) {
	    obj = null;
	    showMessage("");
	}

	if (obj != null) {
	    String message = obj.name + ": " + obj.coords.toString();
	    if (!wasPlanet)
		message += "   mag. " + new Float(obj.mag).toString();

	    showMessage(message);
	}

	return true;
    }

    // Constellation names -- typed in gradually as I get around to it
    String constellations[][] = {
	{ "Sagittarius", "18:45", "-33" },
	{ "Scorpius", "16:40", "-31" },
	{ "Aquila", "20:40", "8" },
	{ "Lyra", "18:45", "35" },
	{ "Cygnus", "20:30", "35" },
	{ "Hercules", "17", "34" },
	{ "Andromeda", "50", "36" },
	{ "Orion", "6", "2" },
	{ "Gemini", "7:30", "25" },
	{ "Ophiuchus", "17:15", "2" },
	{ "Ursa Major", "14:30", "52" },
    };


    // Short list of bright nebulae and other nonstellar objects:
    String nebulae[][] = {
	{ "M1", "5:34:30", "22:2", "8.4" },
	{ "M2", "21:33:17", "0:47", "6.3" },
	{ "M3", "13:42:9", "28:21", "6.4" },
	{ "M4", "16:23:18", "-26:30", "6.4" },
	{ "M5", "15:19:16", "2:5", "6.2" },
	{ "M6", "17:39:38", "-32:13", "5.3" },
	{ "M7", "17:53:40", "-34:48", "3.2" },
	{ "M8", "18:3:32", "-24:23", "6" },
	{ "M9", "17:18:28", "-18:31", "7.3" },
	{ "M10", "16:57:19", "-4:7", "6.7" },
	{ "M11", "18:51:21", "-6:16", "6.3" },
	{ "M12", "16:47:18", "-1:58", "6.6" },
	{ "M13", "16:41:54", "36:27", "5.7" },
	{ "M14", "17:37:19", "-3:15", "7.7" },
	{ "M15", "21:30:13", "12:12", "6" },
	{ "M16", "18:19:25", "-13:47", "6.5" },
	{ "M17", "18:21:27", "-16:11", "7" },
	{ "M18", "18:19:27", "-17:8", "7.5" },
	{ "M19", "17:2:33", "-26:15", "6.6" },
	{ "M20", "18:2:31", "-23:2", "9" },
	{ "M21", "18:4:31", "-22:30", "6.5" },
	{ "M22", "18:36:32", "-23:55", "5.9" },
	{ "M23", "17:56:29", "-19:1", "6.9" },
	{ "M24", "18:18:28", "-18:24", "4.6" },
	{ "M25", "18:32:29", "-19:15", "6" },
	{ "M26", "18:45:22", "-9:24", "9.3" },
	{ "M27", "20:0:5", "22:44", "7.6" },
	{ "M28", "18:24:33", "-24:52", "7.3" },
	{ "M29", "20:23:55", "38:32", "7.1" },
	{ "M30", "21:40:25", "-23:8", "8.4" },
	{ "M31", "0:43:22", "41:17", "4.8" },
	{ "M32", "0:43:22", "40:53", "8.7" },
	{ "M33", "1:34:25", "30:41", "6.7" },
	{ "M34", "2:42:37", "42:49", "5.5" },
	{ "M35", "6:9:32", "24:21", "5.3" },
	{ "M36", "5:36:40", "34:6", "6.3" },
	{ "M37", "5:53:38", "32:33", "6.2" },
	{ "M38", "5:28:41", "35:49", "7.4" },
	{ "M39", "21:32:54", "48:28", "5.2" },
	{ "M40", "12:35:19", "58:13", "9" },
	{ "M41", "6:47:3", "-20:44", "5" },
	{ "M42", "5:35:14", "-5:23", "5" },
	{ "M43", "5:35:14", "-5:16", "7" },
	{ "M44", "8:40:26", "19:59", "3.7" },
	{ "M45", "3:47:29", "24:8", "1.4" },
	{ "M46", "7:42:9", "-14:50", "6" },
	{ "M47", "7:37:9", "-14:30", "4.5" },
	{ "M48", "8:13:14", "-5:46", "5.3" },
	{ "M49", "12:30:16", "7:59", "8.6" },
	{ "M50", "7:3:12", "-8:20", "6.9" },
	{ "M51", "13:30:3", "47:10", "8.1" },
	{ "M52", "23:24:7", "61:37", "7.3" },
	{ "M53", "13:13:14", "18:9", "7.6" },
	{ "M54", "18:54:36", "-30:29", "8.7" },
	{ "M55", "19:39:35", "-30:57", "7.1" },
	{ "M56", "19:16:58", "30:11", "8.2" },
	{ "M57", "18:53:56", "33:3", "9.3" },
	{ "M58", "12:38:16", "11:48", "9.2" },
	{ "M59", "12:42:16", "11:39", "9.6" },
	{ "M60", "12:44:16", "11:33", "9.7" },
	{ "M61", "12:22:17", "4:28", "9.99" },
	{ "M62", "17:1:36", "-30:7", "6.6" },
	{ "M63", "13:16:7", "42:0", "9.5" },
	{ "M64", "12:57:14", "21:40", "8.8" },
	{ "M65", "11:19:18", "13:5", "9.3" },
	{ "M66", "11:20:18", "12:59", "8.4" },
	{ "M67", "8:50:22", "11:48", "6.1" },
	{ "M68", "12:39:20", "-26:45", "8.2" },
	{ "M69", "18:30:38", "-32:21", "8.9" },
	{ "M70", "18:42:38", "-32:19", "9.6" },
	{ "M71", "19:54:7", "18:47", "9" },
	{ "M72", "20:53:22", "-12:33", "9.8" },
	{ "M73", "20:57:22", "-12:43", "9.99" },
	{ "M74", "1:37:21", "15:49", "9.99" },
	{ "M75", "20:5:29", "-21:57", "8" },
	{ "M76", "1:42:35", "51:36", "9.99" },
	{ "M77", "2:43:17", "0:2", "8.9" },
	{ "M78", "5:47:17", "0:3", "8" },
	{ "M79", "5:24:2", "-24:32", "8.4" },
	{ "M80", "16:17:30", "-22:59", "7.7" },
	{ "M81", "9:56:3", "69:2", "7.9" },
	{ "M82", "9:56:5", "69:40", "8.8" },
	{ "M83", "13:37:25", "-29:54", "9.99" },
	{ "M84", "12:25:16", "12:52", "9.3" },
	{ "M85", "12:25:16", "18:10", "9.3" },
	{ "M86", "12:26:16", "12:55", "9.7" },
	{ "M87", "12:31:16", "12:22", "9.2" },
	{ "M88", "12:32:16", "14:24", "9.99" },
	{ "M89", "12:36:16", "12:32", "9.5" },
	{ "M90", "12:37:16", "13:8", "9.97" },
	{ "M91", "12:36:16", "13:54", "9.99" },
	{ "M92", "17:17:46", "43:8", "6.1" },
	{ "M93", "7:44:4", "-23:52", "6" },
	{ "M94", "12:51:11", "41:6", "7.9" },
	{ "M95", "10:43:19", "11:44", "9.99" },
	{ "M96", "10:47:19", "11:48", "9.1" },
	{ "M97", "11:15:27", "54:59", "9.99" },
	{ "M98", "12:14:16", "14:53", "9.99" },
	{ "M99", "12:19:16", "14:24", "9.99" },
	{ "M100", "12:23:16", "15:48", "9.99" },
	{ "M101", "14:2:53", "54:20", "9.6" },
	{ "M102", "15:6:27", "55:45", "9.99" },
	{ "M103", "1:33:40", "60:43", "7.4" },
	{ "M104", "12:40:18", "-11:37", "8.7" },
	{ "M105", "10:48:19", "12:34", "9.2" },
	{ "M106", "12:19:14", "47:17", "8.6" },
	{ "M107", "16:32:24", "-13:5", "9.2" },
	{ "M108", "11:12:27", "55:39", "9.9" },
	{ "M109", "11:58:18", "53:21", "9.9" },
	{ "m110", "0:40:22", "41:43", "9.4" },
	{ "UCNGC104", "0:24:8", "-72:5", "5" },
	{ "NGC225", "0:43:32", "61:48", "6.5" },
	{ "NGC362", "1:2:25", "-70:50", "6" },
	{ "NGC457", "1:19:10", "58:19", "6.5" },
	{ "NGC663", "1:45:57", "61:15", "6.5" },
	{ "NGCGC869", "2:20:45", "57:9", "5.2" },
	{ "NGC2070", "5:38:44", "-69:7", "5.7" },
	{ "NGC2244", "6:32:40", "4:52", "5.7" },
	{ "NGC2440", "7:41:50", "-18:12", "7" },
	{ "NGC2506", "7:59:54", "-10:35", "7.5" },
	{ "NGC2808", "9:11:59", "-64:51", "7" },
	{ "NGC3242", "10:24:43", "-18:38", "7" },
	{ "NGC3532", "11:6:26", "-58:40", "7" },
	{ "NGC3766", "11:36:14", "-61:37", "6.5" },
	{ "NGC4258", "12:18:59", "47:17", "7.5" },
	{ "NGC4565", "12:36:24", "25:59", "7" },
	{ "NGC5322", "13:49:35", "60:10", "6" },
	{ "NGC6025", "16:3:41", "-60:29", "6" },
	{ "NGC6067", "16:13:21", "-54:13", "7" },
	{ "NGC6210", "16:44:28", "25:28", "7" },
	{ "NGC6543", "17:58:35", "66:38", "6.5" },
	{ "NGC6572", "18:12:38", "6:51", "8" },
	{ "NGC6752", "19:10:50", "-59:59", "7" },
	{ "NGC7009", "21:4:8", "-11:22", "7" },
	{ "NGC7243", "22:15:5", "49:53", "6.5" },
	{ "NGC7662", "23:25:49", "42:29", "7" },
	{ "NGC7772", "23:51:33", "16:16", "5" },
	{ "NGC7789", "23:57:2", "56:43", "8" },
    };
}
